// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       rasterized_font.cc
 *  \brief      ラスタライズフォント
 *  \date       Since: June 22, 2021. 18:45:02 JST.
 *  \author     Acerola
 *  \version    1.1.0
 */
/* ------------------------------------------------------------------------- */

#include "mglext/rasterized_font.h"

namespace MGLExt
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 *  \param[in]  path    フォントファイル一式があるディレクトリのパス
 *  \param[in]  name    フォント名
 */
/* ------------------------------------------------------------------------- */
RasterizedFont::RasterizedFont(const char *path, const char *name) noexcept
    : MGL::Render::IndexedFontResource(MGL::Render::FontOrigin::BaseLine)
    , _isValid(false)
    , _fontInfo()
    , _textures()
    , _lineAdvance(0.0f)
    , _rubyOffset(0.0f)
    , _faces()
{
    _isValid = Load(path, name);    
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      フォントの読み込み
 *  \param[in]  path    フォントファイル一式があるディレクトリのパス
 *  \param[in]  name    フォント名
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool RasterizedFont::Load(const char *path, const char *name) noexcept
{
    // パスの末尾に'/'が付いていない場合は付与
    MGL::STL::string directory(path);
    if (directory.back() != '/')
    {
        directory += '/';
    }

    // メトリクス情報を読み込み
    try
    {
        MGL::STL::string metricsFilePath = directory + name + "_metrics.bin";
        MGL::File::ThrowingHandle handle(metricsFilePath.c_str());
        
        // フォント情報を読み込み
        handle.Read(&_fontInfo, sizeof(BinaryHeader));

        // ヘッダの識別子をチェック
        if (_fontInfo.identify != MGL::Hash::FNV1a("FontMetricsV2"))
        {
            MGL_ERROR("[RasterizedFont] %s is not metrics file.", metricsFilePath.c_str());
            return false;
        }
        
        // リビジョンチェック
        if (_fontInfo.revision != 0)
        {
            MGL_ERROR("[RasterizedFont] Metrics revision %s is not supported.", _fontInfo.revision);
            return false;
        }
        
        // テクスチャの読み込み
        for (uint32_t i = 0; i < _fontInfo.textureCount; i++)
        {
            MGL::Render::Texture texture;
            
            constexpr int kPathSize = 512;
            char texturePath[kPathSize];
            if (snprintf(texturePath, kPathSize, "%s%s_%02d.png", directory.c_str(), name, i) >= kPathSize)
            {
                texturePath[kPathSize - 1] = '\0';
            }
            
            texture.Load(texturePath);
            if (!texture)
            {
                MGL_ERROR("[RasterizedFont] Failed to load texture: %s", texturePath);
                return false;
            }
            
            _textures.push_back(texture);
        }

        // 各チャンクを読み込み
        while (!handle.IsEOF())
        {
            BinaryChunkHeader chunkHeader = {};
            handle.Read(&chunkHeader, sizeof(BinaryChunkHeader));
            
            // フェイスチャンク
            if (chunkHeader.identify == MGL::Hash::FNV1a("FaceChunk"))
            {
                // フェイス情報を読み取り
                FaceInfo faceInfo = {};
                handle.Read(&faceInfo, sizeof(FaceInfo));
             
                // マップにフェイスを新規作成
                auto &face = _faces[MGL::Render::FontFaceType{faceInfo.faceType}];
                face.info = faceInfo;
                face.glyphs = MGL::STL::make_unique<MGL::Render::FontGlyph[]>(face.info.glyphCount);
                
                // グリフ情報を読み込み
                auto binaryGlyphs = MGL::STL::make_unique<GlyphInfo[]>(face.info.glyphCount);
                handle.Read(binaryGlyphs.get(), sizeof(GlyphInfo) * face.info.glyphCount);
                
                // 読み込んだグリフ情報からテーブルを構築
                for (size_t i = 0; i < face.info.glyphCount; i++)
                {
                    const auto &binGlyph = binaryGlyphs[i];
                    auto &glyph = face.glyphs[i];

                    glyph.code = binGlyph.code;
                    glyph.size = MGL::Vector2(float(binGlyph.width), float(binGlyph.height));
                    glyph.advance = float(binGlyph.advance);
                    glyph.bearing = MGL::Vector2(float(binGlyph.bearingX), float(binGlyph.bearingY));

                    if (binGlyph.texturePage < _fontInfo.textureCount)
                    {
                        glyph.texture = MGL::Render::TextureWithBounds(
                                _textures[binGlyph.texturePage],
                                MGL::Rectangle(
                                    float(binGlyph.textureX),
                                    float(binGlyph.textureY),
                                    float(binGlyph.textureWidth),
                                    float(binGlyph.textureHeight)));
                    }
                    else
                    {
                        glyph.texture = {};
                    }
                }
            }
            // 未対応のチャンクはスキップ
            else
            {
                handle.Skip(chunkHeader.chunkSize);
            }
        }
    }
    catch ([[maybe_unused]] MGL::File::Result result)
    {
        MGL_ERROR("[RasterizedFont] Failed to load metrics file: %d", result.GetErrorCode());
        return false;
    }
    
    // 行の送りサイズをグリフの高さの最大値に設定
    _lineAdvance = float(_fontInfo.maxHeight);

    // ルビがあるなら追加で色々設定
    if (const auto &m = _faces.find(MGL::Render::FontFaceType::Ruby); m != _faces.end())
    {
        _rubyOffset = _lineAdvance;
        _lineAdvance += float(m->second.info.maxHeight);
        SetEnabled(MGL::Render::FontFeature::Ruby, true);
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      グリフのインデックスを取得
 *  \param[in]  character   取得するUTF-32文字
 *  \param[in]  faceType    フェイスタイプ
 *  \return     対応したグリフのインデックス．見つからない場合は SIZE_MAX
 */
/* ------------------------------------------------------------------------- */
size_t RasterizedFont::GetIndex(char32_t character, MGL::Render::FontFaceType faceType) const noexcept
{
    const auto &m = _faces.find(faceType);
    if (m == _faces.end())
    {
        return SIZE_MAX;
    }
    auto &face = m->second;
    
    size_t min = 0;
    size_t max = face.info.glyphCount;
    
    do
    {
        size_t index = min + (max - min) / 2;
        auto value = face.glyphs[index].code;
        
        if (value == character)
        {
            return index;
        }
        else if (value > character)
        {
            max = index;
        }
        else if (value < character)
        {
            min = index + 1;
        }
        
    } while(min != max);
    
    return SIZE_MAX;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      グリフを取得
 *  \param[in]  index       取得するグリフのインデックス
 *  \param[in]  faceType    フェイスタイプ
 *  \param[in]  option      描画オプション
 *  \retval     対応するグリフ情報．見つからない場合はnullptr
 */
/* ------------------------------------------------------------------------- */
const MGL::Render::FontGlyph *RasterizedFont::GetGlyph(
        size_t index,
        MGL::Render::FontFaceType faceType,
        [[maybe_unused]] const MGL::Render::FontOption &option) const noexcept
{
    const auto &m = _faces.find(faceType);
    if (m == _faces.end())
    {
        return nullptr;
    }
    auto &face = m->second;

    if (index >= face.info.glyphCount)
    {
        return nullptr;
    }

    return &face.glyphs[index];
}
}   // namespace MGLExt
// vim: et ts=4 sw=4 sts=4
